/********************************************************************
* Copyright (c) 1992 InterSoft Co.  All Rights Reserved
* THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF InterSoft Co.
* The copyright notice above does not evidence any
* actual or intended publication of such soucone code.
*
* $Id: yacc.y,v 1.1.1.1 1996/06/06 16:11:29 hernan Exp $
*
* DESCRIPTION
*********************************************************************/

%{

#include <ideafix.h>
#include <ifound/instruct.h>
#include <ifound/stdc.h>
#include <stdio.h>

#if defined(YYBISON) && (defined(SHIDEAIX) || defined(SHIDEAES))
#	if __GNUC__ > 1		/* GNU C and GNU C++ define this.  */
#		define 	__builtin_memcpy(TO,FROM,COUNT)	memcpy(TO,FROM,COUNT)
#	endif
#endif

#define LENGTH	1024
#define NO_PROTOTYPE

static char line[LENGTH] = {NULL};
static char lexeme[LENGTH] = {NULL};
static char *ptr = NULL;
static int in_nflds = 0;

static int exprpar_lex(void);
static int exprpar_error(char *msg);

void codeString(const char *str);
void codeNum(const char *str);
void codeInstruction(enum jInstruction i);
void codeFieldVal(const char *str);
void codeIdentifier(const char *str);
void codeGetEnv(const char *str);
void codeIn(Int nflds);
void codeVoid();
void codeFunction(const char *func);
void codeMsg(const char *tab, const char *sec, const char *msg);
bool isMsgTab(const char *id);

%}

%union {
	int ival;
	char *strval;
};

%token <ival> T_AND T_BETWEEN
%token <ival> T_HOUR T_IN T_NOT T_OR
%token <ival> T_THIS T_TODAY T_TRUE T_FALSE T_NULL T_VOID
%token <strval> T_NUMBER T_ID T_STRING T_GETENV T_MSGTABLE
%token <ival> T_ERROR

%right T_WHEN
%right '?' ':'
%left T_OR 
%left T_AND 
%left T_NOT
%left T_EQ T_NE T_LT T_LE T_GT T_GE T_IN T_BETWEEN
%left '+' '-'
%left '*' '/'
%left unary_min
%left '('
%left '.'

%{
static struct tagSymtab {
	char *word;
	int  token;
} symtab[] = {
	{ "and",		T_AND },
	{ "between",	T_BETWEEN },
	{ "false",		T_FALSE },
	{ "hour", 		T_HOUR },
	{ "in",			T_IN },
	{ "not",		T_NOT },
	{ "null",		T_NULL },
	{ "or",			T_OR },
	{ "this",		T_THIS },
	{ "today",		T_TODAY },
	{ "true",		T_TRUE },
	{ "void",		T_VOID },
	{ "when",		T_WHEN },
};

static int  symtab_size = (sizeof(symtab)/sizeof(symtab[0]));

%}

%%

expr		: when_expr
			| val_expr
			| bool_expr
			;

val_expr	: val_expr '+' val_expr
			{
				codeInstruction(INST_ADD);
			}
			| val_expr '-' val_expr
			{
				codeInstruction(INST_DIF);
			}
			| val_expr '*' val_expr
			{
				codeInstruction(INST_MULT);
			}
			| val_expr '/' val_expr
			{
				codeInstruction(INST_DIV);
			}
			| '-' val_expr %prec unary_min
			{
				codeInstruction(INST_UMIN);
			}		
			| func_begin opt_arg_list func_end
			{
			}
			| message
			{
			}
			| T_NUMBER
			{
				codeNum($1);
			}
			| T_STRING
			{
				codeString($1);
			}
			| T_HOUR
			{
				codeInstruction(INST_HOUR);
			}
			| T_TODAY 
			{
				codeInstruction(INST_TODAY);
			}
			| T_GETENV
			{
				codeGetEnv($1);
			}
			| T_TRUE
			{
				codeNum("1");
			}
			| T_FALSE
			{ 
				codeNum("0");
			}
			| T_NULL
			{
				codeString("");
			}
			| T_ID
			{
				codeIdentifier($1);
			}
			| T_THIS
			{
			    /* NULL_STRING = THIS */
				codeFieldVal("");
			}
			| bool_expr '?' val_expr ':' val_expr
			{
				codeInstruction(INST_IF);
			}
			| bool_expr '?' val_expr 
			{
				codeVoid();
				codeInstruction(INST_IF);
			}
			| T_VOID
			{
				codeVoid();
			}
			| '(' when_expr ')'
			{
			}
			| '(' val_expr ')'
			{
			}
			;

bool_expr	: bool_expr T_AND bool_expr
			{
				codeInstruction(INST_AND);
			}
			| bool_expr T_OR bool_expr
			{
				codeInstruction(INST_OR);
			}
			| T_NOT bool_expr
			{
				codeInstruction(INST_NOT);
			}
			| val_expr T_EQ val_expr
			{
				codeInstruction(INST_EQ);
			}
			| val_expr T_NE val_expr
			{
				codeInstruction(INST_NE);
			}
			| val_expr T_LT val_expr
			{
				codeInstruction(INST_LT);
			}
			| val_expr T_LE val_expr
			{
				codeInstruction(INST_LE);
			}
			| val_expr T_GT val_expr
			{
				codeInstruction(INST_GT);
			} 
			| val_expr T_GE val_expr
			{
				codeInstruction(INST_GE);
			}
			| val_expr T_IN '(' in_list ')'
			{     
				codeIn(in_nflds);	
				in_nflds = 0;
			}
			| val_expr T_BETWEEN val_expr T_AND val_expr
			{
				codeInstruction(INST_BETWEEN);
			}
			| '(' bool_expr ')'
			{
			}
			;

when_expr	: val_expr T_WHEN bool_expr val_expr
			{
				codeInstruction(INST_WHEN);
			}
			| val_expr T_WHEN bool_expr when_expr
			{
				codeInstruction(INST_WHEN);
			}
			| val_expr T_WHEN bool_expr
			{
				codeVoid();
				codeInstruction(INST_WHEN);
			}
			;

in_list		: val_expr
			{
				in_nflds++;
			}
			| in_list ',' val_expr
			{
				in_nflds++;
			}
			;

func_begin	: T_ID '('
			{
				codeFunction($1);
			}	
			;

func_end	: ')'
			{
				codeInstruction(INST_ARGEND);
			}
			;

message		: msg
			{
				codeInstruction(INST_ARGEND);
			}
			| msg '(' opt_arg_list ')'
			{
				codeInstruction(INST_ARGEND);
			}
			;

msg			: T_MSGTABLE '(' T_STRING ')'
			{
				codeMsg($1, NULL, $3);
			}
			| T_MSGTABLE '(' T_STRING ',' T_STRING ')'
			{
				codeMsg($1, $3, $5);
			}
			;

last_arg	: val_expr
			{
			}
			;

arg			: val_expr
			{
				codeInstruction(INST_ARGSEPAR);
			}
			;

arg_list	: last_arg
			{
			}
			| arg ',' arg_list
			{
			}
			;

opt_arg_list : // Empty
			| arg_list
			;
%%

static int getpunt(void)
{
	char *aux = lexeme;
	
	switch(*aux++ = *ptr) {
	case '>' : 
		if (*++ptr == '=') {
			*aux++ = *ptr++;
			*aux++ = '\0'; 
			return T_GE;
		} else {
			*aux++ = '\0'; 
			return T_GT;				
		}
	case '<' : 
		if (*++ptr == '=') {
			*aux++ = *ptr++;
			*aux++ = '\0'; 
			return T_LE;
		} else {
			*aux++ = '\0'; 
			return T_LT;
		}
	case '=' : 
		if (*++ptr == '=') {
			*aux++ = *ptr++;
		}
		*aux++ = '\0'; 
		return T_EQ;
	case '!' :
		if (*++ptr == '=') {
			*aux++ = *ptr++;
			*aux++ = '\0'; 
			return T_NE;
    	}
		return T_ERROR;
	case ':' : 
	case '+' :
	case '-' :
	case '*' :
	case '/' : 
	case '(' : 
	case ')' : 
	case '.' : 
	case ',' : 
	case '?' : 
		*aux++ = '\0';
		return *ptr++;
	}
	return EOF;
}

static int cmprut(char *s1, struct tagSymtab *s2)
{
	return strcmp(s1, s2->word);
}

void setExprLine(const char *s)
{
	assert(strlen(s) < LENGTH);
	strcpy(line, s);
	ptr = line;
}

static char *build_str(char *s)
{
	char *buff = (char *) malloc(strlen(s)+1);
	return strcpy(buff, s);
}

static void skipComments()
{
	int c_ant = 0, c = 0;
	while (((c = *++ptr) != '/' || c_ant != '*') && c != '\0')
		c_ant = c;
	if (c != '\0') ptr++;
	for (; isspace(*ptr); ptr++)
		;
}

static int exprpar_lex(void)
{
	/* skip all whitespaces */
	for (; isspace(*ptr); ptr++)
		;

	/* skip comments */
	if (*ptr == '/') {
		if (*++ptr == '*')
			skipComments();
		else
			ptr--;
	}

	/* is end of line */
	if (*ptr == '\0')
		return EOF;

	/* enviroment variable ? */
	if (*ptr == '$') {
		char *aux;

		/* find end of line or next whitespace */
		for (aux = lexeme, ptr++; isalnum(*ptr) || *ptr == '_'; aux++, ptr++)
			*aux = *ptr;

		*aux = '\0';

		yylval.strval = build_str(lexeme);

		return T_GETENV;		

	} 
	
	/* is a number */
	if (isdigit(*ptr)) {
		char *aux;
		char *pdecpoint = NULL;

		/* find end of this number */
		for (aux = lexeme; isdigit(*ptr) || 
						   (!pdecpoint && *ptr == '.'); aux++, ptr++)
			if ((*aux = *ptr) == '.')
				pdecpoint = ptr;
						
		*aux = '\0';

		yylval.strval = build_str(lexeme);

		return T_NUMBER;
		
	} 
	
	/* is a string */
	if (*ptr == '\"' || *ptr == '\'') {
		char *aux;

		/* find end of this string */
		for (aux = lexeme, ptr++; *ptr && *ptr != '\"' && *ptr != '\''; 
												aux++, ptr++)
			*aux = *ptr;
	
		*aux = '\0';

		if (*ptr)
			ptr++;
			
		yylval.strval = build_str(lexeme);

		return T_STRING;
		
	} 
	
	/* other tokens */
	{
    	char *aux;
		int i, token;

		// find puntuations signs
		if ((token = getpunt()) != EOF)
			return token;
		
		// find end of line or next whitespace
		for (aux = lexeme; isalnum(*ptr) || *ptr == '_'; aux++, ptr++)
			*aux = *ptr;

		*aux = '\0';

		// reserved word?
		if ((i = locate(lexeme, (const char *) symtab, symtab_size, 
				sizeof(symtab[0]), IFPCPCP(cmprut))) >= 0)
			return symtab[i].token;
		
		for (; isspace(*ptr); ptr++)
			;

		yylval.strval = build_str(lexeme);

		return isMsgTab(lexeme) ? T_MSGTABLE : T_ID;
	}                 
}

static int exprpar_error(char *)
{
	return 0;
}


